-- Steve v1.12 by Wumbo
-- Original graphics/sounds by Mojang Studios; I don't own anything
-- This add-on is public domain

freeslot(
	"MT_MCITEM",
	"S_MCITEM",
	"MT_MCRAY",
	"S_MCRAY",
	"sfx_mcoof1",
	"sfx_mcoof2",
	"sfx_mcoof3",
	"sfx_mcmne1",
	"sfx_mcmne2",
	"sfx_mcbrk1",
	"sfx_mcbrk2",
	"sfx_mcbrk3",
	"sfx_mcbrk4",
	"sfx_mcpop1",
	"sfx_mcpop2",
	"sfx_mcpop3"
)

local stevesize = (FRACUNIT*7)/8

local facehelper = {
	createwide = function(frames)
		local faces = {}
		
		for i = 1, 4 do
			local face = P_SpawnMobj(0, 0, 0, MT_MCFACE)
			face.origframe = frames[i]
			face.renderflags = RF_SLOPESPLAT | RF_NOSPLATBILLBOARD | RF_FLOORSPRITE
			face.spriteyoffset = -4*FRACUNIT
			P_CreateFloorSpriteSlope(face)
			faces[i] = face
		end
		
		return faces
	end,
	createlong = function(frames)
		local faces = {}
		
		for i = 1, 2 do
			local face = P_SpawnMobj(0, 0, 0, MT_MCFACE)
			face.origframe = frames[i]
			face.renderflags = RF_PAPERSPRITE
			faces[i] = face
		end
		
		return faces
	end
}

local cuboidhelper = {
	create = function(frames, xsize, ysize, zsize, xpivot, ypivot, zpivot, scale, mirror, rollmode, x, y, z, xyangle, zangle)
		local cuboid = {
			widefaces = facehelper.createwide({unpack(frames, 1, 4)}),
			longfaces = facehelper.createlong({unpack(frames, 5, 6)}),
			
			frames = frames,
			
			xsize = xsize,
			ysize = ysize,
			zsize = zsize,
			xpivot = xpivot,
			ypivot = ypivot,
			zpivot = zpivot,
			
			scale = scale,
			mirror = mirror,
			rollmode = rollmode,
			
			x = x,
			y = y,
			z = z,
			xyangle = xyangle,
			zangle = zangle
		}
		
		return cuboid
	end,
	
	update = function(cuboid, realOnly)
		cuboid.real = {
			x = cuboid.x,
			y = cuboid.y,
			z = cuboid.z,
			xyangle = cuboid.xyangle + ANGLE_180,
			zangle = -cuboid.zangle,
			xsize = FixedMul(cuboid.scale, cuboid.xsize),
			ysize = FixedMul(cuboid.scale, cuboid.ysize),
			zsize = FixedMul(cuboid.scale, cuboid.zsize),
			xpivot = cuboid.xpivot,
			ypivot = -cuboid.ypivot
		}
		
		local sign = cuboid.mirror and -1 or 1
		
		if cuboid.rollmode then
			cuboid.real.xyangle = $ - ANGLE_90 * sign
			cuboid.real.zangle = $ * sign
			cuboid.real.xsize, cuboid.real.ysize = $2, $1
			cuboid.real.xpivot, cuboid.real.ypivot = $2 * sign, -$1 * sign
		end
		
		for i = 0, 3 do
			if cuboid.real.zangle > ANGLE_90 * i and cuboid.real.zangle < ANGLE_90 * i + 256 then
				cuboid.real.zangle = ANGLE_90 * i
				break
			end
		end
		
		if cuboid.real.xpivot != 0 then
			cuboid.real.x = $ + FixedMul(sin(cuboid.real.xyangle), -cuboid.real.xpivot)
			cuboid.real.y = $ + FixedMul(cos(cuboid.real.xyangle), cuboid.real.xpivot)
		end
		if cuboid.real.ypivot != 0 then
			cuboid.real.x = $ + FixedMul(cos(cuboid.real.xyangle), cuboid.real.ypivot)
			cuboid.real.y = $ + FixedMul(sin(cuboid.real.xyangle), cuboid.real.ypivot)
		end
		if cuboid.zpivot != 0 then
			cuboid.real.x = $ + FixedMul(sin(cuboid.real.zangle), FixedMul(cos(cuboid.real.xyangle), cuboid.zpivot))
			cuboid.real.y = $ + FixedMul(sin(cuboid.real.zangle), FixedMul(sin(cuboid.real.xyangle), cuboid.zpivot))
			cuboid.real.z = $ + FixedMul(cos(cuboid.real.zangle), -cuboid.zpivot)
		end
		
		if realOnly then return end
		
		for _, faces in ipairs({cuboid.widefaces, cuboid.longfaces}) do
			for _, face in ipairs(faces) do face.frame = face.origframe end
		end
		
		if cuboid.rollmode then
			cuboid.widefaces[1].frame, cuboid.widefaces[3].frame, cuboid.longfaces[1].frame, cuboid.longfaces[2].frame =  $4, $3, $1, $2
		end
		
		for i, face in ipairs(cuboid.widefaces) do
			face.frame = $ & ~FF_PAPERSPRITE | FF_FLOORSPRITE
			if i == 2 or i == 3 then face.frame = $ ^^ FF_VERTICALFLIP end
			if i == 3 then face.frame = $ ^^ FF_HORIZONTALFLIP end
		end
		for _, face in ipairs(cuboid.longfaces) do
			face.frame = $ & ~FF_FLOORSPRITE | FF_PAPERSPRITE
		end
		
		local zanglefixed = AngleFixed(cuboid.real.zangle)
		for i, face in ipairs(cuboid.widefaces) do
			local widezangle = cuboid.real.zangle + ANGLE_90 * i
			local widezanglefixed = AngleFixed(widezangle)
			
			if widezangle == ANGLE_90 or widezangle == ANGLE_270 then
				face.floorspriteslope.xydirection = cuboid.real.xyangle
				face.floorspriteslope.zangle = ANGLE_90 - 1
				if widezangle == ANGLE_270 then face.frame = $ ^^ FF_VERTICALFLIP end
			elseif widezanglefixed > 270*FRACUNIT then
				face.floorspriteslope.xydirection = cuboid.real.xyangle + ANGLE_180
				face.floorspriteslope.zangle = -widezangle
			elseif widezanglefixed > 180*FRACUNIT then
				face.floorspriteslope.xydirection = cuboid.real.xyangle
				face.floorspriteslope.zangle = abs(ANGLE_180 - widezangle)
				face.frame = $ ^^ FF_VERTICALFLIP
			elseif widezanglefixed > 90*FRACUNIT then
				face.floorspriteslope.xydirection = cuboid.real.xyangle + ANGLE_180
				face.floorspriteslope.zangle = abs(widezangle - ANGLE_180)
				face.frame = $ ^^ FF_VERTICALFLIP
			else
				face.floorspriteslope.xydirection = cuboid.real.xyangle
				face.floorspriteslope.zangle = widezangle
			end
			
			face.angle = cuboid.real.xyangle
			face.spritexscale = cuboid.scale
			face.spriteyscale = FixedMul(cos(face.floorspriteslope.zangle), cuboid.scale + (face.floorspriteslope.zangle == ANGLE_90 - 1 and 1024 or 0))
			
			if cuboid.mirror then face.frame = $ ^^ FF_HORIZONTALFLIP end
			if cuboid.rollmode and i % 2 == 0 then
				face.angle = $ + ANGLE_90 * sign
				face.spritexscale, face.spriteyscale = face.spriteyscale, face.spritexscale
			end
			
			local wideoffset = (i % 2 == 1 and cuboid.real.ysize or cuboid.real.zsize) / 2
			P_MoveOrigin(face,
				cuboid.real.x + FixedMul(cos(cuboid.real.xyangle), FixedMul(sin(widezangle), -wideoffset)),
				cuboid.real.y + FixedMul(sin(cuboid.real.xyangle), FixedMul(sin(widezangle), -wideoffset)),
				cuboid.real.z + FixedMul(cos(widezangle),  wideoffset) + FixedMul(4*FRACUNIT, cuboid.scale)
			)
			
			face.floorspriteslope.o = {face.x, face.y, face.z}
		end
		
		local function adjustpos(center) return FixedMul(cuboid.scale, max(0, min(FRACUNIT, -abs((zanglefixed - center*FRACUNIT) / 90) + (3*FRACUNIT)/2))) end
		local longxyoffset = cuboid.real.xsize / 2
		local longxyangle = cuboid.real.xyangle + ANGLE_90
		local longxyadjust = adjustpos(cuboid.mirror and 225 or 135) * sign
		local longzadjust = adjustpos(cuboid.mirror and 135 or 225)
		for i, face in ipairs(cuboid.longfaces) do
			face.angle = cuboid.real.xyangle
			if cuboid.mirror then face.frame = $ ^^ FF_HORIZONTALFLIP end
			face.spritexscale = cuboid.scale
			face.spriteyscale = cuboid.scale
			face.spriteroll = cuboid.real.zangle * sign
			
			local sign = (i * 2 - 3) * sign
			P_MoveOrigin(face,
				cuboid.real.x + FixedMul(longxyoffset, cos(longxyangle)) * sign + FixedMul(longxyadjust, cos(cuboid.real.xyangle)),
				cuboid.real.y + FixedMul(longxyoffset, sin(longxyangle)) * sign + FixedMul(longxyadjust, sin(cuboid.real.xyangle)),
				cuboid.real.z + longzadjust
			)
		end
	end,
	
	destroy = function(cuboid)
		for _, faces in ipairs({cuboid.widefaces, cuboid.longfaces}) do
			for _, face in ipairs(faces) do
				if face.valid then P_RemoveMobj(face) end
			end
		end
		cuboid = nil
	end,
	
	modifywide = function(cuboid, modifier) for _, face in ipairs(cuboid.widefaces) do modifier(face) end end,
	modifylong = function(cuboid, modifier) for _, face in ipairs(cuboid.longfaces) do modifier(face) end end,
	
	validate = function(cuboid)
		for _, faces in ipairs({cuboid.widefaces, cuboid.longfaces}) do
			for _, face in ipairs(faces) do
				if not face.valid then return false end
			end
		end
		return true
	end
}

local stevehelper = {
	create = function(player, scale)
		if player.steve == nil then player.steve = {} end
		local steve = player.steve
		
		if steve.model == nil then
			local argslist = {
				{{A, B, C, D, E, F}, 32*FRACUNIT, 32*FRACUNIT, 32*FRACUNIT, 0, 0, 0, 0, false, false, 0, 0, 0, 0, 0},
				{{G, H, I, J, K, L}, 32*FRACUNIT, 16*FRACUNIT, 48*FRACUNIT, 0, 0, 0, 0, false, false, 0, 0, 0, 0, 0},
				{{M, N, O, P, Q, R}, 16*FRACUNIT, 16*FRACUNIT, 48*FRACUNIT, 0, 0, 0, 0, true,  false, 0, 0, 0, 0, 0},
				{{M, N, O, P, Q, R}, 16*FRACUNIT, 16*FRACUNIT, 48*FRACUNIT, 0, 0, 0, 0, false, false, 0, 0, 0, 0, 0},
				{{S, T, U, V, W, X}, 16*FRACUNIT, 16*FRACUNIT, 48*FRACUNIT, 0, 0, 0, 0, true,  false, 0, 0, 0, 0, 0},
				{{S, T, U, V, W, X}, 16*FRACUNIT, 16*FRACUNIT, 48*FRACUNIT, 0, 0, 0, 0, false, false, 0, 0, 0, 0, 0},
			}
			
			steve.model = {
				main = {},
				mask = {},
				argslist = argslist,
				
				x = player.mo.x,
				y = player.mo.y,
				z = player.mo.z,
				xyangle = player.mo.angle,
				zangle = player.aiming,
				
				bodyxyangle = player.mo.angle,
				bodyzangle = 0,
				scale = scale
			}
			
			for i, args in ipairs(steve.model.argslist) do
				steve.model.main[i] = cuboidhelper.create(unpack(args))
				local modifier = function(face) face.dontdrawforviewmobj = player.mo end
				cuboidhelper.modifywide(steve.model.main[i], modifier)
				cuboidhelper.modifylong(steve.model.main[i], modifier)
			end
		end
		
		if steve.play == nil then
			steve.play = {
				moving = false,
				jumping = false,
				sneaking = false,
				sprinting = false,
				bhopping = false,
				swimming = false,
				sitting = false,
				hanging = false,
				tossing = false,
				
				underwater = false,
				goopwater = false,
				submerged = false,
				antigravity = false,
				
				dying = false,
				
				hurttimer = 0,
				redtimer = 0,
				blinktimer = 0,
				
				breath = 20,
				breathoffset = 0,
				poptimer = 0,
				
				usepos = 0,
				usemax = 15,
				usetimer = 0,
				
				punchaction = false,
				punchactive = false,
				placeaction = false,
				placeactive = false,
				
				soundtimer = 0,
				
				sprintfov = 8*FRACUNIT,
				currentfov = 0,
				
				swing = {
					progress = 0,
					progressadd = ANG20,
					extent = 0,
					realextent = 0,
					extentadd = FixedAngle(4*FRACUNIT),
					minextent = ANG20,
					maxextent = ANG60
				},
				bob = {
					progress = 0,
					progressadd = FixedAngle(12*FRACUNIT),
					extent = 0,
					realextent = 0,
					extentadd = FRACUNIT/8,
					minextent = 0,
					maxextent = 2*FRACUNIT
				},
				
				last = {}
			}
		end
		
		if steve.persist == nil then
			steve.persist = {
				health = 20,
				hotbar = {},
				lasthotbar = {},
				hotbarslot = 0
			}
			
			for i = 0, 8 do
				steve.persist.hotbar[i] = {item = MT_NULL, amount = 0}
				steve.persist.lasthotbar[i] = {item = MT_NULL, amount = 0}
			end
		end
	end,
	
	update = function(player)
		local steve = player.steve
		local mainmodel, maskmodel = steve.model.main, steve.model.mask
		local realscale = FixedMul(steve.model.scale, player.mo.scale)
		
		for i, cuboid in ipairs(mainmodel) do
			local modifier = function(face)
				face.color = player.skincolor
				if player.mo.flags2 & MF2_DONTDRAW and steve.play.redtimer == 0 then
					face.flags2 = $ | MF2_DONTDRAW
				else
					face.flags2 = $ & ~MF2_DONTDRAW
				end
			end
			cuboidhelper.modifywide(cuboid, modifier)
			cuboidhelper.modifylong(cuboid, modifier)
			
			cuboid.xpivot, cuboid.ypivot, cuboid.zpivot = 0, 0, 0
			cuboid.scale = realscale + (i != 2 and realscale / 4096 or 0)
			cuboid.rollmode = false
			
			cuboid.x, cuboid.y, cuboid.z = steve.model.x, steve.model.y, steve.model.z
			cuboid.xyangle, cuboid.zangle = steve.model.bodyxyangle, 0
		end
		
		mainmodel[1].zpivot = realscale * -16
		mainmodel[1].z = $ + realscale * 96
		
		mainmodel[2].z = $ + realscale * 72
		
		for i = 3, 4 do
			local sign = i * 2 - 7
			mainmodel[i].xpivot = realscale * (24 * sign)
			mainmodel[i].zpivot = realscale * 16
			mainmodel[i].z = $ + realscale * 88
		end
		
		for i = 5, 6 do
			local sign = i * 2 - 11
			mainmodel[i].xpivot = realscale * (8 * sign)
			mainmodel[i].zpivot = realscale * 24
			mainmodel[i].z = $ + realscale * 48
		end
		
		if not steve.play.dying then
			local xybobstate = FixedMul(sin(steve.play.bob.progress), FixedMul(player.mo.scale, steve.play.bob.extent))
			local zbobstate = FixedMul(-abs(cos(steve.play.bob.progress)), FixedMul(player.mo.scale, steve.play.bob.extent))
			for i, cuboid in ipairs(mainmodel) do
				cuboid.x = $ + FixedMul(cos(steve.model.xyangle + ANGLE_90), xybobstate)
				cuboid.y = $ + FixedMul(sin(steve.model.xyangle + ANGLE_90), xybobstate)
				cuboid.z = $ + zbobstate - (steve.play.sitting and realscale * 36 or 0)
			end
			
			mainmodel[1].xyangle = steve.model.xyangle
			mainmodel[1].zangle = steve.model.zangle * (steve.play.antigravity and -1 or 1)
			
			local armswingstate = FixedMul(sin(steve.play.swing.progress), steve.play.swing.extent)
			for i = 3, 4 do
				local sign = i * 2 - 7
				if steve.play.sitting then
					mainmodel[i].zangle = ANG30
				elseif steve.play.hanging then
					mainmodel[i].zangle = ANGLE_180
				else
					mainmodel[i].zangle = armswingstate * -sign
				end
			end
			
			local legswingstate = (armswingstate / 4) * 5
			for i = 5, 6 do
				local sign = i * 2 - 11
				if steve.play.sitting or steve.play.hanging then
					mainmodel[i].xyangle = $ + ANG15 * -sign
					mainmodel[i].zangle = FixedAngle(80*FRACUNIT)
				else
					mainmodel[i].zangle = legswingstate * sign
				end
			end
			
			if steve.play.sneaking then
				mainmodel[1].z = $ + realscale * -10
				
				mainmodel[2].ypivot = realscale * -8
				mainmodel[2].z = $ + realscale * -8
				mainmodel[2].zangle = ANGLE_337h
				
				for i = 3, 4 do
					mainmodel[i].z = $ + realscale * -8
					if not steve.play.sitting and not steve.play.hanging then
						mainmodel[i].zangle = $ - ANG15
					end
				end
				
				for i = 5, 6 do
					mainmodel[i].ypivot = $ + realscale * -12
				end
			end
			
			if steve.play.usepos > 0 and steve.play.usepos < steve.play.usemax then
				local arm = mainmodel[4]
				
				local initxyangle, initzangle = arm.xyangle, arm.zangle
				arm.xyangle = steve.model.bodyxyangle + (steve.play.antigravity and ANGLE_180 or 0)
				arm.zangle = 0
				cuboidhelper.update(arm, true)
				
				arm.xpivot = 0
				arm.x = arm.real.x
				arm.y = arm.real.y
				arm.xyangle = $ + (steve.play.antigravity and ANGLE_180 or 0)
				
				local progress = FixedDiv(steve.play.usepos*FRACUNIT, steve.play.usemax*FRACUNIT)
				if progress < FRACUNIT/5 then
					local segprogress = progress * 5
					arm.xyangle = $ + ease.linear(segprogress, -ANG60, 0)
					arm.zangle = ease.outsine(segprogress, initzangle, ANGLE_90)
				elseif progress < (FRACUNIT*2)/5 then
					local segprogress = (progress - FRACUNIT/5) * 5
					arm.xyangle = $ + ease.outsine(segprogress, 0, ANG60)
					arm.zangle = ease.insine(segprogress, ANGLE_90, ANGLE_45)
				else
					local segprogress = FixedDiv(progress - (FRACUNIT*2)/5, (FRACUNIT*3)/5)
					arm.xyangle = $ + ease.outsine(segprogress, ANG60, 0)
					arm.zangle = ease.outsine(segprogress, ANGLE_45, initzangle)
				end
			end
		else
			for _, cuboid in ipairs(mainmodel) do cuboid.rollmode = true end
			steve.play.redtimer = -1
			steve.model.bodyzangle = min(ANGLE_90, $ + max(ANG1, (ANGLE_90 - $) / 7))
			if steve.play.antigravity then
				steve.model.z = player.mo.z + realscale * 120
			end
		end
		
		local realbodyzangle = steve.model.bodyzangle + (steve.play.antigravity and ANGLE_180 or 0)
		local modelbackup = {}
		if realbodyzangle != 0 then
			for i, cuboid in ipairs(mainmodel) do
				local x, y, z = cuboid.x, cuboid.y, cuboid.z
				local xpivot, ypivot, zpivot = cuboid.xpivot, cuboid.ypivot, cuboid.zpivot
				local xyangle, zangle = cuboid.xyangle, cuboid.zangle
				
				cuboid.xpivot, cuboid.ypivot, cuboid.zpivot = 0, 0, 0
				cuboid.xyangle = steve.model.bodyxyangle + ((steve.play.antigravity and not cuboid.rollmode) and ANGLE_180 or 0)
				cuboidhelper.update(cuboid, true)
				
				cuboid.z = steve.model.z + ((not steve.play.dying) and realscale * 64 or 0)
				cuboid.zpivot = cuboid.z - cuboid.real.z
				cuboid.zangle = realbodyzangle
				cuboidhelper.update(cuboid, true)
				
				cuboid.x, cuboid.y, cuboid.z = cuboid.real.x, cuboid.real.y, cuboid.real.z
				if cuboid.rollmode then
					cuboid.z = $ + FixedMul(sin(cuboid.zangle), xpivot)
				else
					cuboid.z = $ + FixedMul(sin(cuboid.zangle), ypivot)
				end
				
				cuboid.xpivot, cuboid.ypivot, cuboid.zpivot = xpivot, ypivot, zpivot
				if cuboid.rollmode then
					cuboid.xpivot = FixedMul(cos(cuboid.zangle), $)
				else
					cuboid.ypivot = FixedMul(cos(cuboid.zangle), $)
				end
				
				cuboid.xyangle = $ + (steve.model.bodyxyangle - xyangle) * ((i == 1 and steve.play.antigravity) and -1 or 1)
				cuboid.zangle = $ + zangle
				cuboidhelper.update(cuboid, false)
				
				modelbackup[i] = {x, y, z, xpivot, ypivot, xyangle, zangle}
			end
		else
			for _, cuboid in ipairs(mainmodel) do cuboidhelper.update(cuboid, false) end
		end
		
		if steve.play.redtimer > 0 or steve.play.redtimer < 0 then
			if #maskmodel == 0 then
				for i, args in ipairs(steve.model.argslist) do
					maskmodel[i] = cuboidhelper.create(unpack(args))
					local modifier = function(face)
						face.origframe = $ | FF_TRANS70
						face.color = SKINCOLOR_RED
						face.colorized = true
						face.dontdrawforviewmobj = player.mo
					end
					cuboidhelper.modifywide(maskmodel[i], modifier)
					cuboidhelper.modifylong(maskmodel[i], modifier)
				end
			else
				for _, cuboid in ipairs(maskmodel) do
					local modifier = function(face) face.flags2 = $ & ~MF2_DONTDRAW end
					cuboidhelper.modifywide(cuboid, modifier)
					cuboidhelper.modifylong(cuboid, modifier)
				end
			end
			for i, cuboid in ipairs(maskmodel) do
				for key, value in pairs(mainmodel[i]) do
					if key == "longfaces" or key == "widefaces" or key == "frames" then continue end
					cuboid[key] = value
				end
				cuboidhelper.update(cuboid, false)
			end
		elseif #maskmodel != 0 then
			for _, cuboid in ipairs(maskmodel) do
				local modifier = function(face) face.flags2 = $ | MF2_DONTDRAW end
				cuboidhelper.modifywide(cuboid, modifier)
				cuboidhelper.modifylong(cuboid, modifier)
			end
		end
		
		if realbodyzangle != 0 then
			for i, cuboid in ipairs(mainmodel) do
				cuboid.x, cuboid.y, cuboid.z, cuboid.xpivot, cuboid.ypivot, cuboid.xyangle, cuboid.zangle = unpack(modelbackup[i])
			end
		end
	end,
	
	/*
		Restore the player to a non-Steveified state
		
		Modes:
		0 - Destroy model
		1 - The above + wipe non-persistent player information
		2 - The above + reset health
		3 - The above + wipe persistent player information
		4 - The above + revert player to default state
	*/
	destroy = function(player, mode)
		local steve = player.steve
		if steve == nil then return end
		
		-- Remove Steve model
		if steve.model != nil then
			for _, model in ipairs({steve.model.main, steve.model.mask}) do
				for _, cuboid in ipairs(model) do cuboidhelper.destroy(cuboid) end
			end
			steve.model = nil
		end
		if mode == 0 then return end
		
		-- Remove non-persistent player information
		steve.play = nil
		if mode == 1 then return end
		
		-- Reset health
		if steve.persist != nil then steve.persist.health = 20 end
		if mode == 2 then return end
		
		-- Remove persistent player information
		steve.persist = nil
		if mode == 3 then return end
		
		-- Remove Steve object
		steve = nil
		
		-- Re-enable vanilla HUD
		hud.enable("score")
		hud.enable("time")
		hud.enable("rings")
		hud.enable("lives")
		
		-- Reset speed values
		player.normalspeed = skins[player.mo.skin].normalspeed
		player.acceleration = skins[player.mo.skin].acceleration
	end,
	
	-- Ensure that a player is Steve and is treated as such
	validinfo = function(player)
		return player.mo != nil and player.mo.skin == "steve" and player.steve != nil and player.steve.play != nil and player.steve.persist != nil
	end,
	
	-- Ensure that a player has a valid Steve model
	validmodel = function(player)
		for _, model in ipairs({player.steve.model.main, player.steve.model.mask}) do
			for _, cuboid in ipairs(model) do
				if not cuboidhelper.validate(cuboid) then return false end
			end
		end
		return true
	end,
	
	-- Get movement speed relative to maximum speed
	currentspeed = function(player)
		local currentspeed = FixedDiv(FixedHypot(player.rmomx, player.rmomy), player.mo.scale)
		if player.powers[pw_sneakers] or player.powers[pw_super] then
			currentspeed = ($ * 3) / 5
		end
		if P_IsObjectOnGround(player.mo) then
			currentspeed = FixedDiv($, player.mo.friction)
			if player.mo.eflags & MFE_UNDERWATER then
				currentspeed = $ * 2
			end
		end
		return currentspeed
	end,
	
	-- Return maximum desired speed based on current state
	maxspeed = function(player)
		local speed = skins[player.mo.skin].normalspeed
		if player.steve.play.sneaking   then speed = $ / 2 end
		if player.steve.play.sprinting  then speed = ($ * 4) / 3 end
		if player.steve.play.bhopping   then speed = ($ * 5) / 4 end
		if player.steve.play.underwater then speed = ($ * 3) / 2 end
		return speed
	end,
	
	-- Deal a specific amount of damage to the player
	damage = function(player, source, amount)
		local steve = player.steve
		
		-- Take health away from player
		steve.persist.health = $ - amount
		
		-- Set timers for invincibility, red tint, and health bar blink
		steve.play.hurttimer = 19
		steve.play.redtimer = 10
		steve.play.blinktimer = 15
		
		-- Throw the player back if the damage is caused by another object, and launch upwards if on the ground
		if source != nil and source.valid then
			local throwangle = R_PointToAngle2(player.mo.x, player.mo.y, source.x, source.y) + ANGLE_180
			local gravsign = steve.play.antigravity and -1 or 1
			P_InstaThrust(player.mo, throwangle, player.mo.scale * 7)
			if P_IsObjectOnGround(player.mo) then
				player.mo.z = $ + gravsign
				player.mo.momz = player.mo.scale * 10 * gravsign
			elseif player.mo.momz < 0 then
				player.mo.momz = 0
			end
		end
		
		if steve.persist.health <= 0 then
			-- Kill player if there is no health left
			P_DamageMobj(player.mo, nil, nil, 0, DMG_INSTAKILL)
		else
			-- Play "oof" sound if player is not dead yet
			-- Otherwise the sound is played by the death handler
			local mcoof = {sfx_mcoof1, sfx_mcoof2, sfx_mcoof3}
			S_StartSound(player.mo, mcoof[P_RandomKey(#mcoof) + 1])
		end
	end,
	
	-- Spawn an item object travelling at a certain angle and velocity
	dropitem = function(source, type, xyangle, zangle, momentum, antigravity)
		local dropheight = 0
		if source.type == MT_PLAYER then
			local steve = source.player.steve
			if not antigravity then
				dropheight = source.z + FixedMul(steve.model.scale, source.scale) * 96
			else
				dropheight = source.z + source.height - FixedMul(steve.model.scale, source.scale) * 96
			end
		else
			dropheight = source.z + source.height / 2
		end
		
		local item = P_SpawnMobj(source.x, source.y, dropheight, MT_MCITEM)
		local scale = FixedDiv(24*FRACUNIT, max(mobjinfo[type].radius * 2, mobjinfo[type].height))
		
		item.momx = FixedMul(momentum, FixedMul(cos(xyangle), cos(zangle)))
		item.momy = FixedMul(momentum, FixedMul(sin(xyangle), cos(zangle)))
		item.momz = FixedMul(momentum, (sin(zangle) * 2) / 3)
		
		item.sprite = states[mobjinfo[type].spawnstate].sprite
		item.frame = states[mobjinfo[type].spawnstate].frame
		
		item.radius = mobjinfo[type].radius
		item.height = mobjinfo[type].height
		
		item.renderflags = RF_PAPERSPRITE
		item.spritexscale = scale
		item.spriteyscale = scale
		
		item.angle = source.angle + ANGLE_90
		item.friction = FRACUNIT/4
		
		if antigravity then
			item.flags2 = $ | MF2_OBJECTFLIP
			if source.type != MT_PLAYER then item.momz = -$ end
		end
		
		item.realtype = type
		item.tosstimer = source.type == MT_PLAYER and 56 or 14
	end,
	
	-- Backup contents of hotbar
	backuphotbar = function(player)
		local hotbar = player.steve.persist.hotbar
		local lasthotbar = player.steve.persist.lasthotbar
		for slot = 0, 8 do
			lasthotbar[slot].item = hotbar[slot].item
			lasthotbar[slot].amount = hotbar[slot].amount
		end
	end,
	
	-- Restore contents of hotbar from backup
	restorehotbar = function(player)
		local hotbar = player.steve.persist.hotbar
		local lasthotbar = player.steve.persist.lasthotbar
		for slot = 0, 8 do
			hotbar[slot].item = lasthotbar[slot].item
			hotbar[slot].amount = lasthotbar[slot].amount
		end
	end,
	
	// Remove all items from hotbar
	clearhotbar = function(player)
		for slot = 0, 8 do
			player.steve.persist.hotbar[slot].amount = 0
		end
	end,
	
	/*
		Below is a glossary of variables for this stupid dumb animation function that is
		tailored specifically for back-and-forth animations, of which there are two of them
		
		progress		position of the animation in angle form
		progressadd		progress is incremented by this, multiplied by modifier
		extent			current allowed variation of the animation from its starting point;
						unless immediate is true, value is updated only when progress reaches
						a multiple of 180 degrees (equivalent to starting point)
		realextent		the animation's "real" variation, used to update extent when needed
		extendadd		realextent is incremented by this, or decremented if moving is false
		minextent		minimum allowed variation of the animation;
						used when the animation has just started, but not when it is ending
		maxextent		maximum allowed variation of the animation;
						multiplied by modifier only if modifier is below FRACUNIT, to allow
						fine-grained control of the variation while keeping it in-bounds
		
		animation		the animation being updated
		modifier		progressadd (and maxextent if below FRACUNIT) is multiplied by this
		moving			if true, animation progresses normally;
						if false, extent is steadily decreased to zero, stopping the animation
		immediate		if true, extent always has the same value as realextent
		
		The current variation will be decremented by double the value of extentadd if/while it
		exceeds the maximum variation, in order to quickly correct it to the new maximum
	*/
	animate = function(animation, modifier, moving, immediate)
		local progressadd = FixedMul(animation.progressadd, modifier)
		local maxextent = animation.maxextent
		local extentadd = animation.extentadd
		
		if modifier < FRACUNIT then
			maxextent = min($, FixedMul($, modifier))
		end
		
		if animation.extent > maxextent then
			moving = false
			extentadd = $ * 2
		end
		
		if moving then
			animation.realextent = min(maxextent, max(animation.minextent, $) + extentadd)
			
			local oldsign = animation.progress != 0 and animation.progress / abs(animation.progress) or 0
			animation.progress = $ + progressadd
			local newsign = animation.progress != 0 and animation.progress / abs(animation.progress) or 0
			
			if oldsign != newsign or immediate then animation.extent = animation.realextent end
		elseif animation.extent > 0 then
			animation.realextent = max(0, $ - extentadd)
			animation.progress = $ + progressadd
			animation.extent = animation.realextent
		else
			animation.realextent = 0
			animation.progress = 0
			animation.extent = 0
		end
	end
}

-- Give the player an item
stevehelper.giveitem = function(player, type, droponfull)
	if not stevehelper.validinfo(player) then return end
	
	local hotbar = player.steve.persist.hotbar
	for pass = 1, 2 do
		for slot = 0, 8 do
			-- Try to add item to an existing stack during the first pass
			-- Otherwise, try to add item to an empty slot during the second pass
			if (pass == 1 and (hotbar[slot].amount <= 0 or hotbar[slot].item == MT_NULL))
			or (pass == 2 and hotbar[slot].amount > 0 and hotbar[slot].item != MT_NULL) then
				continue
			end
			
			if hotbar[slot].amount <= 0 or hotbar[slot].item == MT_NULL
			or (hotbar[slot].amount < 64 and hotbar[slot].item == type) then
				hotbar[slot].amount = hotbar[slot].item == MT_NULL and 1 or max(1, $ + 1)
				hotbar[slot].item = type
				return true
			end
		end
	end
	
	if droponfull then
		stevehelper.dropitem(player.mo, type, player.mo.angle, player.aiming, 20*FRACUNIT, player.steve.play.antigravity)
	end
	
	return false
end

-- Replace vanilla HUD with a variation of Minecraft's HUD
stevehelper.customhud = function(screen, player)
	if stevehelper.validinfo(player) then
		-- Disable vanilla HUD
		hud.disable("score")
		hud.disable("time")
		hud.disable("rings")
		hud.disable("lives")
		
		-- Cache normal and blinking heart graphics
		local fullheart = {
			screen.cachePatch("MCHTFULA"),
			screen.cachePatch("MCHTFULB")
		}
		local halfheart = {
			screen.cachePatch("MCHTHLFA"),
			screen.cachePatch("MCHTHLFB"),
		}
		local noneheart = {
			screen.cachePatch("MCHTNONA"),
			screen.cachePatch("MCHTNONB")
		}
		
		local steve = player.steve
		
		-- Render health bar
		local health = steve.persist.health
		local blinktimer = steve.play.blinktimer
		local mode = (blinktimer > 0 and (blinktimer / 4) % 2 == 0) and 2 or 1
		for heart = 0, 9 do
			local heart2 = heart * 2
			local x = 69 + heart * 9
			if heart2 < health then
				if heart2 == health - 1 then
					screen.draw(x, 168, halfheart[mode], V_SNAPTOBOTTOM)
				else
					screen.draw(x, 168, fullheart[mode], V_SNAPTOBOTTOM)
				end
			else
				screen.draw(x, 168, noneheart[mode], V_SNAPTOBOTTOM)
			end
		end
		
		if steve.play.submerged then
			-- Cache normal and popping bubble graphics
			local normalbubble = screen.cachePatch("MCBUFULL")
			local poppedbubble = screen.cachePatch("MCBUPOPD")
			
			-- Render breath bubbles
			local breath = (steve.play.breath + (steve.play.breath % 2)) / 2
			local poptimer = steve.play.poptimer
			for bubble = 1, 10 do
				local x = 161 + (10 - bubble) * 9
				if bubble <= breath then
					screen.draw(x, 168, normalbubble, V_SNAPTOBOTTOM)
				elseif steve.play.breath % 2 == 0 and bubble == breath + 1 and poptimer > 0 then
					screen.draw(x, 168, poppedbubble, V_SNAPTOBOTTOM)
				end
			end
		end
		
		-- Render hotbar
		local hotbarslot = steve.persist.hotbarslot
		screen.draw(69, 178, screen.cachePatch("MCHBFRNT"), V_SNAPTOBOTTOM)
		screen.draw(72, 181, screen.cachePatch("MCHBBACK"), V_SNAPTOBOTTOM | V_30TRANS)
		screen.draw(68 + hotbarslot * 20, 177, screen.cachePatch("MCHBSLCT"), V_SNAPTOBOTTOM)
		
		local hotbar = steve.persist.hotbar
		for slot = 0, 8 do
			if hotbar[slot].item == MT_NULL or hotbar[slot].amount <= 0 then continue end
			
			-- Render items
			local itemstate = states[mobjinfo[hotbar[slot].item].spawnstate]
			local patch = screen.getSpritePatch(itemstate.sprite, itemstate.frame, 3)
			if patch == nil or not patch.valid then patch = screen.getSpritePatch(SPR_UNKN) end
			local scale = FixedDiv(16*FRACUNIT, max(patch.width, patch.height)*FRACUNIT)
			local xoffset = (16*FRACUNIT - patch.width * scale) / 2
			local yoffset = (16*FRACUNIT - patch.height * scale) / 2
			screen.drawScaled(
				(72 + 20 * slot)*FRACUNIT + patch.leftoffset * scale + xoffset,
				181*FRACUNIT + patch.topoffset * scale + yoffset,
				scale, patch, V_SNAPTOBOTTOM
			)
			
			-- Render quantity
			local numbers = screen.cachePatch("MCHBNUMS")
			local digit1 = hotbar[slot].amount % 10
			local digit2 = ((hotbar[slot].amount - digit1) / 10) % 10
			if digit1 > 1 or digit2 > 0 then
				for i, digit in ipairs({digit1, digit2}) do
					if i > 1 and digit == 0 then continue end
					screen.drawCropped(
						(83 + 6 * (1 - i) + 20 * slot)*FRACUNIT, 190*FRACUNIT, FRACUNIT, FRACUNIT, numbers,
						V_SNAPTOBOTTOM, nil, 6*FRACUNIT * digit, 0, 6*FRACUNIT, 8*FRACUNIT
					)
				end
			end
		end
	end
end
hud.add(stevehelper.customhud, "game")

-- Implement give command
COM_AddCommand("mc_give", function(player, type, amount)
	if type != nil then
		type = string.upper(type)
		if string.sub(type, 1, 3) != "MT_" then
			type = "MT_" .. type
		end
	end
	
	if not stevehelper.validinfo(player) then
		CONS_Printf(player, "This command can only be used while playing as Steve.")
		return
	elseif type == nil or not pcall(function() return _G[type] end) then
		CONS_Printf(player, "mc_give <objtype> [<amount>]: give yourself an item")
		return
	end
	
	for i = 1, (amount == nil and 1 or max(1, amount)) do
		stevehelper.giveitem(player, _G[type], true)
	end
end, COM_ADMIN)

-- Implement clear command
COM_AddCommand("mc_clear", function(player)
	if not stevehelper.validinfo(player) then
		CONS_Printf(player, "This command can only be used while playing as Steve.")
		return
	end
	stevehelper.clearhotbar(player)
end, COM_ADMIN)

-- Implement heal command
COM_AddCommand("mc_heal", function(player)
	if not stevehelper.validinfo(player) then
		CONS_Printf(player, "This command can only be used while playing as Steve.")
		return
	end
	player.steve.persist.health = 20
end, COM_ADMIN)

-- Implement backup command
COM_AddCommand("mc_backup", function(player)
	if not stevehelper.validinfo(player) then
		CONS_Printf(player, "This command can only be used while playing as Steve.")
		return
	end
	stevehelper.backuphotbar(player)
end, COM_ADMIN)

-- Implement restore command
COM_AddCommand("mc_restore", function(player)
	if not stevehelper.validinfo(player) then
		CONS_Printf(player, "This command can only be used while playing as Steve.")
		return
	end
	stevehelper.restorehotbar(player)
end, COM_ADMIN)

-- Implement variable to toggle whether health persists between levels
CV_RegisterVar({"mc_healthpersist", "Off", CV_NETVAR | CV_CHEAT, CV_OnOff})

-- Implement variable to toggle whether items persist between levels
CV_RegisterVar({"mc_itempersist", "On", CV_NETVAR | CV_CHEAT, CV_OnOff})

-- Implement variable to toggle whether items persist after dying
-- In singleplayer, hotbar contents are restored to their state at the last checkpoint
-- In multiplayer, hotbar contents are kept exactly the same
CV_RegisterVar({"mc_keepinventory", "On", CV_NETVAR | CV_CHEAT, CV_OnOff})

addHook("PreThinkFrame", do
	for player in players.iterate do
		if not stevehelper.validinfo(player) then continue end
		local steve = player.steve
		local last = steve.play.last
		
		-- Decrement timers each frame
		if steve.play.hurttimer > 0 then steve.play.hurttimer = $ - 1 end
		if steve.play.redtimer > 0 then steve.play.redtimer = $ - 1 end
		if steve.play.blinktimer > 0 then steve.play.blinktimer = $ - 1 end
		if steve.play.usetimer > 0 then steve.play.usetimer = $ - 1 end
		if steve.play.soundtimer > 0 then steve.play.soundtimer = $ - 1 end
		
		-- Stop all movement if dead, and drop items if keepinventory is not enabled
		if steve.play.dying then
			if CV_FindVar("mc_keepinventory").value == 0 then
				local hotbar = steve.persist.hotbar
				for slot = 0, 8 do
					if hotbar[slot].item != MT_NULL and hotbar[slot].amount > 0 then
						for i = 1, hotbar[slot].amount do
							stevehelper.dropitem(player.mo, hotbar[slot].item, FixedAngle(((360*FRACUNIT) / 9) * slot), ANGLE_45, 8*FRACUNIT, steve.play.antigravity)
						end
						hotbar[slot].amount = 0
					end
				end
			end
			continue
		end
		
		-- Decrement remaining breath if underwater
		if player.mo.state == S_PLAY_GASP then
			steve.play.breath = 20
			steve.play.breathoffset = leveltime % 35
		elseif steve.play.submerged and player.exiting == 0 and not (player.pflags & PF_GODMODE) then
			player.powers[pw_underwater] = 1050
			if leveltime % 35 == steve.play.breathoffset then
				if steve.play.breath >= 0 then
					steve.play.breath = $ - 1
					steve.play.poptimer = 5
				else
					stevehelper.damage(player, nil, 2)
				end
			end
		end
		if steve.play.poptimer > 0 then steve.play.poptimer = $ - 1 end
		
		-- Change hotbar slot
		if player.cmd.buttons & BT_WEAPONNEXT then
			steve.persist.hotbarslot = ($ + 1) % 9
		elseif player.cmd.buttons & BT_WEAPONPREV then
			steve.persist.hotbarslot = ($ + 8) % 9
		end
		
		-- Toggle moving state
		steve.play.moving = player.cmd.forwardmove != 0 or player.cmd.sidemove != 0
		
		-- Toggle jumping state and repeat jump while button is held
		if (player.cmd.buttons & BT_JUMP) and last.airborne and P_IsObjectOnGround(player.mo) then
			player.cmd.buttons = $ & ~BT_JUMP
			steve.play.jumping = true
		elseif steve.play.jumping then
			player.cmd.buttons = $ | BT_JUMP
			steve.play.jumping = false
		end
		
		-- Toggle sneaking state
		if not steve.play.sneaking and (player.cmd.buttons & BT_CUSTOM1) and not (steve.play.sitting or steve.play.hanging) then
			steve.play.sneaking = true
		elseif (steve.play.sneaking and not (player.cmd.buttons & BT_CUSTOM1)) or steve.play.sitting or steve.play.hanging then
			steve.play.sneaking = false
		end
		
		-- Toggle sprinting state
		local currentspeed = FixedDiv(player.speed, player.mo.scale)
		if not steve.play.sneaking and not steve.play.sprinting and (player.cmd.buttons & BT_SPIN)
		and player.cmd.forwardmove > 0 and currentspeed > (skins[player.mo.skin].normalspeed * 2) / 3 then
			steve.play.sprinting = true
		elseif steve.play.sneaking or steve.play.sprinting and (player.cmd.forwardmove <= 0
		or currentspeed < (skins[player.mo.skin].normalspeed * 2) / 3) then
			steve.play.sprinting = false
		end
		
		-- Toggle bunny hopping state
		if player.pflags & PF_JUMPED and steve.play.sprinting then
			steve.play.bhopping = true
		elseif not (player.pflags & PF_JUMPED) and steve.play.bhopping then
			steve.play.bhopping = false
		end
		
		-- Toggle sitting and hanging states
		steve.play.sitting = player.powers[pw_carry] == CR_MACESPIN or player.powers[pw_carry] == CR_MINECART
		steve.play.hanging = player.powers[pw_carry] == CR_PLAYER or player.powers[pw_carry] == CR_ROPEHANG
		
		-- Toggle tossing state and toss selected item
		if not steve.play.tossing and (player.cmd.buttons & BT_CUSTOM2) then
			local slot = steve.persist.hotbar[steve.persist.hotbarslot]
			if slot.item != MT_NULL and slot.amount > 0 then
				stevehelper.dropitem(player.mo, slot.item, player.mo.angle, player.aiming, 20*FRACUNIT, steve.play.antigravity)
				slot.amount = $ - 1
			end
			steve.play.tossing = true
		elseif steve.play.tossing and not (player.cmd.buttons & BT_CUSTOM2) then
			steve.play.tossing = false
		end
		
		-- Toggle water states and cushion Z momentum upon entering water
		if not steve.play.underwater and (player.mo.eflags & MFE_UNDERWATER) and not (player.mo.eflags & MFE_GOOWATER) then
			player.mo.momz = $ / 2
			steve.play.underwater = true
		elseif not steve.play.goopwater and (player.mo.eflags & MFE_GOOWATER) then
			player.mo.momz = $ / 2
			steve.play.goopwater = true
		elseif steve.play.underwater and not (player.mo.eflags & MFE_UNDERWATER) then
			steve.play.underwater = false
		elseif steve.play.goopwater and not (player.mo.eflags & MFE_GOOWATER) then
			steve.play.goopwater = false
		end
		
		-- Toggle submerged states and initiate breath countdown
		if not steve.play.submerged and (player.mo.eflags & MFE_UNDERWATER) and not (player.mo.eflags & MFE_TOUCHWATER) then
			steve.play.breath = 20
			steve.play.breathoffset = leveltime % 35
			steve.play.submerged = true
		elseif steve.play.submerged and (not (player.mo.eflags & MFE_UNDERWATER) or (player.mo.eflags & MFE_TOUCHWATER)) then
			steve.play.submerged = false
		end
		
		-- Toggle antigravity state
		if not steve.play.antigravity and (player.mo.eflags & MFE_VERTICALFLIP) then
			steve.play.antigravity = true
		elseif steve.play.antigravity and not (player.mo.eflags & MFE_VERTICALFLIP) then
			steve.play.antigravity = false
		end
		
		-- Increment use animation position
		if steve.play.usepos > 0 and steve.play.usepos < steve.play.usemax then
			steve.play.usepos = ($ + 1) % steve.play.usemax
		end
		
		-- Toggle used states and initiate appropriate action
		if steve.play.usetimer == 0 then
			if not steve.play.punchactive and (player.cmd.buttons & BT_ATTACK) then
				steve.play.usepos = 1
				steve.play.usetimer = 7
				steve.play.punchaction = true
				steve.play.punchactive = true
			elseif steve.play.punchactive and not (player.cmd.buttons & BT_ATTACK) then
				steve.play.punchactive = false
			end
			if not steve.play.placeactive and (player.cmd.buttons & BT_FIRENORMAL) then
				steve.play.placeaction = true
				steve.play.placeactive = true
			elseif steve.play.placeactive and not (player.cmd.buttons & BT_FIRENORMAL) then
				steve.play.placeactive = false
			end
		end
		
		local gravsign = steve.play.antigravity and -1 or 1
		
		-- Swim upwards if jump button is held while underwater
		if steve.play.underwater then
			if (player.cmd.buttons & BT_JUMP) then
				player.mo.momz = min(FixedMul(5*FRACUNIT, player.mo.scale), $ * gravsign + ($ * gravsign < 0 and (FRACUNIT*2)/3 or FRACUNIT/3)) * gravsign
			elseif player.mo.momz * gravsign > 0 then
				player.mo.momz = $ - FixedMul(FRACUNIT/3, player.mo.scale) * gravsign
			end
		end
		
		-- Jump out of water if swimming to surface while moving against a wall
		if not steve.play.underwater and last.underwater and (player.cmd.buttons & BT_JUMP) and player.powers[pw_pushing] then
			player.mo.momz = 15*FRACUNIT * gravsign
		end
		
		if not P_IsObjectOnGround(player.mo) and not (player.mo.eflags & MFE_UNDERWATER) and not player.powers[pw_carry] then
			-- Increase gravitational pull
			player.mo.momz = $ + P_GetMobjGravity(player.mo) * 2
			-- Always jump to maximum height
			if player.pflags & PF_JUMPED then
				player.cmd.buttons = $ | BT_JUMP
			end
		end
		
		-- Increase force of springs, breaking minecarts, and dust devils
		if (player.mo.eflags & MFE_SPRUNG)
		or last.minecart and player.powers[pw_carry] != CR_MINECART
		or last.dustdevil and player.powers[pw_carry] != CR_DUSTDEVIL then
			player.mo.momx = ($ * 7) / 4
			player.mo.momy = ($ * 7) / 4
			player.mo.momz = ($ * 29) / 16
		end
		
		-- Increase force of mace spins
		if last.macespin and player.powers[pw_carry] != CR_MACESPIN then
			player.mo.momx = ($ * 7) / 2
			player.mo.momy = ($ * 7) / 2
			player.mo.momz = ($ * 29) / 16
		end
		
		-- Backup relevant states/properties for use in subsequent tic
		last.x = player.mo.x
		last.y = player.mo.y
		last.momz = player.mo.momz
		last.airborne = not P_IsObjectOnGround(player.mo)
		last.underwater = steve.play.underwater
		last.minecart = player.powers[pw_carry] == CR_MINECART
		last.dustdevil = player.powers[pw_carry] == CR_DUSTDEVIL
		last.macespin = player.powers[pw_carry] == CR_MACESPIN
	end
end)

addHook("ThinkFrame", do
	for player in players.iterate do
		if not stevehelper.validinfo(player) or not stevehelper.validmodel(player) or player.steve.play.dying then continue end
		local steve = player.steve
		local last = steve.play.last
		
		if player.starposttime == leveltime then
			stevehelper.backuphotbar(player)
		end
		
		-- Update position of Steve model
		steve.model.x = player.mo.x
		steve.model.y = player.mo.y
		steve.model.z = player.mo.z
		
		-- Update angle of Steve's head
		steve.model.xyangle = player.mo.angle
		steve.model.zangle = player.aiming
		
		-- Update max speed according to current state
		player.normalspeed = stevehelper.maxspeed(player)
		player.acceleration = stevehelper.maxspeed(player)/FRACUNIT
		
		-- Recreate Minecraft's body turning behavior
		local turned = false
		if abs(player.rmomx) > FRACUNIT/4 or abs(player.rmomy) > FRACUNIT/4 then
			local direction = max(-FRACUNIT, min(FRACUNIT, (sin(R_PointToAngle2(0, 0, player.rmomx, player.rmomy) - steve.model.xyangle) * 3) / 2))
			local destination = steve.model.xyangle + FixedAngle(direction * 45)
			local distance = destination - steve.model.bodyxyangle
			if steve.model.bodyxyangle + distance == destination then
				steve.model.bodyxyangle = $ + distance / 3
				turned = true
			end
		end
		if not turned then
			local distance = steve.model.xyangle - steve.model.bodyxyangle
			if abs(distance) > ANGLE_45 then
				steve.model.bodyxyangle = $ + (abs(distance) - ANGLE_45) * (distance / abs(distance))
			end
		end
		
		-- Construct animation speed modifier
		local modifier = FRACUNIT
		if steve.play.sneaking   then modifier = $ / 4 end
		if steve.play.sprinting  then modifier = ($ * 7) / 6 end
		if steve.play.underwater then modifier = ($ * 2) / 3 end
		
		-- Update progress of bobbing and arm swinging animations
		local airborne = not P_IsObjectOnGround(player.mo) and not (player.pflags & PF_JUMPED)
		if player.speed > 0 and (steve.play.moving and not player.powers[pw_carry]) or airborne then
			if not player.powers[pw_pushing] and P_IsObjectOnGround(player.mo) then
				stevehelper.animate(steve.play.bob, modifier, true, true)
			elseif player.powers[pw_pushing] or airborne then
				modifier = min(FRACUNIT, FixedMul($, FixedDiv(stevehelper.currentspeed(player), player.normalspeed)))
			end
			stevehelper.animate(steve.play.swing, modifier, true, false)
		else
			stevehelper.animate(steve.play.swing, modifier, false, false)
			stevehelper.animate(steve.play.bob, modifier, false, true)
		end
		
		-- Increase FOV while sprinting
		if steve.play.sprinting and not (steve.play.sitting or steve.play.hanging) then
			steve.play.currentfov = min(steve.play.sprintfov, $ + (steve.play.sprintfov - $) / 3)
			player.fovadd = steve.play.currentfov
		elseif steve.play.currentfov > 0 then
			steve.play.currentfov = ($ * 3) / 4
			player.fovadd = steve.play.currentfov
		end
		
		-- Disable first-person camera bob and adjust while sneaking
		local viewz = FixedMul(steve.model.scale, player.mo.scale) * (steve.play.sneaking and 102 or 112)
		player.viewz = player.mo.z + (steve.play.antigravity and player.mo.height or 0) + viewz * (steve.play.antigravity and -1 or 1)
		
		-- Prevent player from moving off of cliff edges while sneaking
		if steve.play.sneaking and P_IsObjectOnGround(player.mo) and last.x != nil and last.y != nil then
			if not steve.play.antigravity then
				if player.mo.floorz - P_FloorzAtPos(player.mo.x, last.y, player.mo.z, player.mo.height) > MAXSTEPMOVE then
					P_MoveOrigin(player.mo, last.x, player.mo.y, player.mo.z)
				end
				if player.mo.floorz - P_FloorzAtPos(last.x, player.mo.y, player.mo.z, player.mo.height) > MAXSTEPMOVE then
					P_MoveOrigin(player.mo, player.mo.x, last.y, player.mo.z)
				end
			else
				if player.mo.ceilingz - P_CeilingzAtPos(player.mo.x, last.y, player.mo.z, player.mo.height) < -MAXSTEPMOVE then
					P_MoveOrigin(player.mo, last.x, player.mo.y, player.mo.z)
				end
				if player.mo.ceilingz - P_CeilingzAtPos(last.x, player.mo.y, player.mo.z, player.mo.height) < -MAXSTEPMOVE then
					P_MoveOrigin(player.mo, player.mo.x, last.y, player.mo.z)
				end
			end
		end
		
		local rayx, rayy, rayz = player.mo.x, player.mo.y, player.viewz - (18*FRACUNIT) * (steve.play.antigravity and -1 or 1)
		local raymomx = FixedMul(player.mo.scale, FixedMul(cos(player.mo.angle), cos(player.aiming)))
		local raymomy = FixedMul(player.mo.scale, FixedMul(sin(player.mo.angle), cos(player.aiming)))
		local raymomz = FixedMul(player.mo.scale, sin(player.aiming))
		
		local ray = P_SpawnMobj(rayx, rayy, rayz, MT_MCRAY)
		ray.momx, ray.momy, ray.momz = raymomx, raymomy, raymomz
		ray.drawonlyforplayer = player
		if steve.play.antigravity then
			ray.renderflags = $ | RF_VERTICALFLIP
		end
		
		local maxdist = 192
		
		-- Cast ray to detect objects that can be hit/mined
		for dist = 1, maxdist do
			P_SceneryXYMovement(ray)
			P_SceneryZMovement(ray)
			
			if ray.collideobj != nil then
				if ray.collideobj.flags & (MF_ENEMY | MF_BOSS) then
					ray.canhit = true
					maxdist = dist
					break
				elseif ray.collideobj.flags & (MF_SOLID | MF_SPRING) then
					ray.canmine = true
					maxdist = dist
					break
				end
			elseif ray.z == ray.floorz then
				for fof in ray.subsector.sector.ffloors() do
					if (fof.flags & FF_EXISTS) and (fof.flags & FF_BUSTUP) and ray.z == fof.topheight then
						ray.bustfof = fof
					end
				end
			end
		end
		
		local slot = steve.persist.hotbar[steve.persist.hotbarslot]
		local empty = slot.item == MT_NULL or slot.amount <= 0
		
		if not empty then
			P_SetOrigin(ray, rayx, rayy, rayz)
			ray.momx, ray.momy, ray.momz = raymomx, raymomy, raymomz
			ray.radius = mobjinfo[slot.item].radius
			ray.height = mobjinfo[slot.item].height
			
			-- Cast ray to detect surfaces that can be placed on
			for dist = 1, maxdist do
				P_SceneryXYMovement(ray)
				P_SceneryZMovement(ray)
				
				if ray.momx != raymomx or ray.momy != raymomy or ray.momz != raymomz then
					ray.canplace = not (
						(mobjinfo[slot.item].flags & MF_SOLID) and R_PointToDist2(rayx, rayy, ray.x, ray.y) <= player.mo.radius * 2
						and ray.z > player.mo.z - ray.height and ray.z < player.mo.z + player.mo.height
					)
					break
				end
			end
		end
		
		local permitplace = not empty and (ray.canmine or ray.canplace)
		
		-- Show "ghost" of object as it would appear if it were placed
		if permitplace then
			ray.sprite = states[mobjinfo[slot.item].spawnstate].sprite
			ray.frame = states[mobjinfo[slot.item].spawnstate].frame | FF_TRANS50
			if steve.play.antigravity then ray.frame = $ | FF_VERTICALFLIP end
			ray.angle = player.mo.angle
		end
		
		-- Always allow player to mine solid objects by holding the punch button
		if steve.play.punchactive and ray.canmine and not steve.play.punchaction then
			steve.play.punchactive = false
		end
		
		-- Always allow player to place objects by holding the place button
		if steve.play.placeactive and ray.canplace and not steve.play.placeaction then 
			steve.play.placeactive = false
		end
		
		-- Damage object if hittable, or mine object if mineable
		if steve.play.punchaction then
			if ray.canhit then
				P_DamageMobj(ray.collideobj, player.mo, player.mo)
			elseif ray.canmine then
				if ray.collideobj.breakage == nil then
					ray.collideobj.breakage = FixedSqrt(FixedMul(ray.collideobj.radius * 2, ray.collideobj.height)) / 5
				else
					ray.collideobj.breakage = $ - FRACUNIT
				end
				if ray.collideobj.breakage <= 0 then
					stevehelper.dropitem(ray.collideobj, ray.collideobj.type, 0, ANGLE_90, 8*FRACUNIT, steve.play.antigravity)
					P_RemoveMobj(ray.collideobj)
					local mcbrk = {sfx_mcbrk1, sfx_mcbrk2, sfx_mcbrk3, sfx_mcbrk4}
					S_StartSound(player.mo, mcbrk[P_RandomKey(#mcbrk) + 1])
				else
					local mcmne = {sfx_mcmne1, sfx_mcmne2}
					S_StartSound(player.mo, mcmne[P_RandomKey(#mcmne) + 1])
				end
			elseif ray.bustfof then
				EV_CrumbleChain(ray.bustfof)
			end
			steve.play.punchaction = false
		end
		
		-- Place held object
		if steve.play.placeaction then
			if permitplace then
				local spawnedobj = P_SpawnMobj(ray.x, ray.y, ray.z, slot.item)
				spawnedobj.angle = player.mo.angle
				spawnedobj.flags = $ | MF_NOGRAVITY
				if steve.play.antigravity then spawnedobj.eflags = $ | MFE_VERTICALFLIP end
				
				slot.amount = $ - 1
				steve.play.usepos = 1
				steve.play.usetimer = 5
				
				local mcbrk = {sfx_mcbrk1, sfx_mcbrk2, sfx_mcbrk3, sfx_mcbrk4}
				S_StartSound(player.mo, mcbrk[P_RandomKey(#mcbrk) + 1])
			end
			steve.play.placeaction = false
		end
	end
end)

addHook("PostThinkFrame", do
	for player in players.iterate do
		if player.mo != nil then
			-- player.skin is used for this check so Steve's items don't disappear when entering NiGHTS special stages
			if skins[player.skin].name == "steve" then
				if not stevehelper.validinfo(player) or not stevehelper.validmodel(player) then
					-- Ensure Steve players have a valid model and properties
					if player.mo.skin == "steve" then
						stevehelper.destroy(player, 0)
						stevehelper.create(player, stevesize)
						-- If the level has just begun, apply effects of persistence variables
						if leveltime == 0 then
							if CV_FindVar("mc_healthpersist").value == 0 then
								player.steve.persist.health = 20
							end
							if CV_FindVar("mc_itempersist").value == 0 then
								stevehelper.clearhotbar(player)
							end
						end
					end
				elseif player.playerstate == PST_REBORN then
					-- Reset model and health if player dies
					stevehelper.destroy(player, 2)
					-- If playing in singleplayer with mc_keepinventory enabled, restore hotbar contents to last checkpoint
					if not netgame and player.steve.persist != nil and CV_FindVar("mc_keepinventory").value == 1 then
						stevehelper.restorehotbar(player)
					end
				else
					-- Update Steve model each frame
					player.mo.sprite2 = SPR2_NULL
					stevehelper.update(player)
				end
			elseif player.steve != nil then
				-- Restore player to default state if skin is changed away from Steve
				stevehelper.destroy(player, 4)
				-- Disable spawnkilling protection if active
				player.mo.flags = $ & ~MF_NOCLIPTHING
			end
		end
	end
end)

addHook("PlayerThink", function(player)
	if player.mo == nil or skins[player.skin].name != "steve" then return end
	
	-- If Steve has just respawned in multiplayer, make him temporarily intangible to objects to prevent spawnkilling
	if netgame and player.powers[pw_flashing] then
		player.mo.flags = $ | MF_NOCLIPTHING
	elseif not player.powers[pw_flashing] then
		player.mo.flags = $ & ~MF_NOCLIPTHING
	end
	
	local map = mapheaderinfo[gamemap]
	
	-- Work around issue where Steve's height causes him to be forced into the shuttle in BCZ3
	if (leveltime == 0 or player.playerstate == PST_REBORN) and map.lvlttl == "Black Core" and map.actnum == 3 then
		player.mo.z = 1920*FRACUNIT
	end
	
	-- Change Steve's height in NiGHTS stages to fix collision and camera angle
	if player.nightstime != 0 then
		player.mo.height = skins[player.mo.skin].height
		player.height = player.mo.height
	elseif player.height != skins[player.skin].height then
		player.height = skins[player.skin].height
	end
end)

-- Regenerate all pre-existing Steve models locally upon joining a netgame
-- Otherwise, any already-existing Steve players will be invisible until they change their skin
addHook("PlayerJoin", function(playernum)
	if #consoleplayer == playernum then
		for player in players.iterate do
			if stevehelper.validinfo(player) then
				stevehelper.destroy(player, 0)
				stevehelper.create(player, stevesize)
			end
		end
	end
end)

-- Destroy Steve model and properties upon quitting netgame
addHook("PlayerQuit", function(player)
	if player.steve != nil then stevehelper.destroy(player, 4) end
end)

-- Synchronize each player's Steve properties
addHook("NetVars", function(network)
	for player in players.iterate do player.steve = network($) end
end)

-- Initiate Minecraft-style damage behavior upon taking damage
addHook("MobjDamage", function(target, inflictor)
	if stevehelper.validinfo(target.player) then
		if target.player.steve.play.hurttimer == 0 then
			if inflictor != nil and inflictor.valid then
				-- Damage player by an amount based on the inflictor's size
				stevehelper.damage(target.player, inflictor, max(1, FixedSqrt(FixedMul(inflictor.radius * 2, inflictor.height) / 200)/FRACUNIT))
			else
				-- If the inflictor isn't an object (ie. lava), deal 1 1/2 hearts of damage
				stevehelper.damage(target.player, nil, 3)
			end
		end
		return true
	end
end, MT_PLAYER)

-- Initiate Minecraft-style death animation upon dying
addHook("MobjDeath", function(target)
	if stevehelper.validinfo(target.player) then
		local player = target.player
		
		player.steve.play.dying = true
		player.mo.momx, player.mo.momy, player.mo.momz = 0, 0, 0
		
		local mcoof = {sfx_mcoof1, sfx_mcoof2, sfx_mcoof3}
		S_StartSound(player.mo, mcoof[P_RandomKey(#mcoof) + 1])
		
		return true
	end
end, MT_PLAYER)

-- Check if ray has collided with an object
addHook("MobjMoveCollide", function(ray, object)
	if ray.collideobj == nil and object.info.doomednum != -1 and ray.z >= object.z and ray.z < object.z + object.height then
		ray.collideobj = object
	end
end, MT_MCRAY)

-- Check if ray has collided with a bustable block
addHook("MobjLineCollide", function(ray, line)
	for _, sector in ipairs({line.frontsector, line.backsector}) do
		for fof in sector.ffloors() do
			if (fof.flags & FF_EXISTS) and (fof.flags & FF_BUSTUP) and ray.z + ray.height > fof.bottomheight and ray.z < fof.topheight then
				ray.bustfof = fof
			end
		end
	end
end, MT_MCRAY)

addHook("MobjThinker", function(item)
	-- Decrement pickup timer each frame
	item.tosstimer = max(0, $ - 1)
	
	-- Continuously rotate item horizontally
	item.angle = $ + ANG1
	
	-- Bob up and down while on the ground
	if P_IsObjectOnGround(item) then
		if item.bobangle == nil then
			item.bobangle = 0
		else
			item.bobangle = $ + ANG1
		end
		item.spriteyoffset = FixedDiv((-cos(item.bobangle * 3) + FRACUNIT) * 10, item.spriteyscale)
	end
	
	-- Move at double speed along the Z axis to simulate Minecraft's gravity
	P_ZMovement(item)
end, MT_MCITEM)

addHook("MobjCollide", function(playerobj, object)
	if not stevehelper.validinfo(playerobj.player) or playerobj.player.steve.play.dying
	or object.z <= playerobj.z - object.height or object.z >= playerobj.z + playerobj.height then
		return
	end
	
	local player = playerobj.player
	if object.type == MT_MCITEM and object.tosstimer == 0 then
		-- Collect item in hotbar
		if stevehelper.giveitem(player, object.realtype, false) then
			P_RemoveMobj(object)
			if player.steve.play.soundtimer == 0 then
				local mcpop = {sfx_mcpop1, sfx_mcpop2, sfx_mcpop3}
				S_StartSoundAtVolume(player.mo, mcpop[P_RandomKey(#mcpop) + 1], 96)
				player.steve.play.soundtimer = 1
			end
		end
	elseif object.type == MT_METALSONIC_BATTLE and object.state == S_METALSONIC_FLOAT then
		-- Force the Metal Sonic boss to spin out upon collision
		-- Because otherwise it just... doesn't happen? For some reason???
		object.flags2 = MF2_CLASSICPUSH
	end
end, MT_PLAYER)